import pytest
from primrose.base.conditional_path_node import AbstractConditionalPath
from primrose.configuration.configuration import Configuration
from primrose.node_factory import NodeFactory

def test_all_nodes_to_prune():
    class TestConditionalNode(AbstractConditionalPath):
        @staticmethod
        def necessary_config(node_config): return set()
        def destinations_to_prune(self):
            return ["csv_writer2"]
        def run(self, data_object):
            return data_object, False

    NodeFactory().register("TestConditionalNode", TestConditionalNode)

    config = {
        "implementation_config": {
            "reader_config": {
                "conditional_node": {
                    "class": "TestConditionalNode",
                    "destinations": ['csv_writer', 'csv_writer2']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                },
                "csv_writer2": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": ["csv_writer3"]
                },
                "csv_writer3": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    node = TestConditionalNode(configuration, "conditional_node")
    assert node.destinations_to_prune() == ["csv_writer2"]
    all_nodes = node.all_nodes_to_prune()
    assert len(all_nodes) == 2
    assert "csv_writer2" in all_nodes
    assert "csv_writer3" in all_nodes

    config = {
        "implementation_config": {
            "reader_config": {
                "conditional_node": {
                    "class": "TestConditionalNode"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    node = TestConditionalNode(configuration, "conditional_node")

    with pytest.raises(Exception) as e:
        node.all_nodes_to_prune()
    assert "Destinations key is missing" in str(e)

def test_all_nodes_to_prune2():
    class TestConditionalNode(AbstractConditionalPath):
        @staticmethod
        def necessary_config(node_config): return set()
        def destinations_to_prune(self):
            return ["junk"]
        def run(self, data_object):
            return data_object, False
    NodeFactory().register("TestConditionalNode", TestConditionalNode)

    config = {
        "implementation_config": {
            "reader_config": {
                "conditional_node": {
                    "class": "TestConditionalNode",
                    "destinations": ['csv_writer']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    node = TestConditionalNode(configuration, "conditional_node")

    with pytest.raises(Exception) as e:
        node.all_nodes_to_prune()
    assert "Destination junk is not in destinations list" in str(e)



"""Module to run a basic regression model

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import logging
from primrose.models.sklearn_model import SklearnModel

class SklearnRegressionModel(SklearnModel):

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys

        Note:
            While you might expect `model` here, we do not need it when in predict or eval mode as the model is cached, only in train

        Returns:
            set of required keys

        """
        return SklearnModel.necessary_config(node_config)

    def fit_training_data(self):
        """fit training data to model"""
        self.model.fit(self.X_train, self.y_train)

    def get_scores(self):
        """get the scores for y_test
        
        Returns:
            dictionary of scores
        
        """
        return SklearnModel.evaluate_regression_metrics(self.y_test, self.predictions)




import pytest
from primrose.node_factory import NodeFactory
from primrose.writers.abstract_file_writer import AbstractFileWriter
from primrose.configuration.configuration import Configuration
from primrose.readers.csv_reader import CsvReader

def test_init():
    f1 = NodeFactory()
    f2 = NodeFactory()

    class TestWriter(AbstractFileWriter):
        def __init__(self, configuration, instance_name):
            pass
        def write(self, data_dict):
            pass

    f1.register("TestWriter", TestWriter)

    n = len(set(f1.name_dict.keys()))
    assert set(f1.name_dict.keys()) == set(f2.name_dict.keys())
    assert f1 != f2
    assert "TestWriter" in f1.name_dict
    assert "TestWriter" in f2.name_dict

    # we are doing this to cleanup the singleton and maintain state
    f1.unregister("TestWriter")

    # but we as well make into a test
    assert len(set(f1.name_dict.keys())) == n - 1
    assert "TestWriter" not in f1.name_dict

def test_unregister():
    f1 = NodeFactory()
    class TestWriter(AbstractFileWriter):
        def __init__(self, configuration, instance_name):
            pass
        def write(self, data_dict):
            pass

    n_before = len(set(f1.name_dict.keys()))
    f1.register("TestWriter", TestWriter)
    n_after = len(set(f1.name_dict.keys()))
    assert n_after == n_before + 1

    f1.unregister('TestWriter')

    assert len(set(f1.name_dict.keys())) == n_before

    with pytest.raises(Exception) as e:
        f1.unregister('doesnotexist')
    assert 'Key not found doesnotexist' in str(e)

def test_valid_configuration():
    f = NodeFactory()
    class TestWriter(AbstractFileWriter):
        def __init__(self, configuration, instance_name):
            super(TestWriter, self).__init__(configuration, instance_name)
        @staticmethod
        def necessary_config(node_config):
            return set(['key1', 'key2'])
        def write(self, data_dict):
            pass

    f.register('TestWriter', TestWriter)

    config = {
        "class": "TestWriter",
        "key1": "someval1",
        "key2": "someval2",
        "destinations": []
    }

    is_valid = f.valid_configuration(TestWriter, config)
    assert is_valid

    missing_config = {
        "class": "TestWriter",
        "key2": "someval2",
        "destinations": []
    }

    with pytest.raises(Exception) as e:
        f.valid_configuration(TestWriter, missing_config)
    assert 'Configuration missing necessary keys for' in str(e)



import pytest
from primrose.configuration.configuration import Configuration
from primrose.readers.csv_reader import CsvReader
from primrose.data_object import DataObject, DataObjectResponseType

def test_init_ok():
    config = {
        "implementation_config":{
            "reader_config": {
            "csv_reader": {
                "class": "CsvReader",
                "filename": "test/minimal.csv",
                "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = CsvReader(configuration, "csv_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    df = data_object.get('csv_reader', rtype=DataObjectResponseType.VALUE.value)
    assert df is not None
    assert df.shape == (2, 2)

    node_config = {
                "class": "CsvReader",
                "filename": "test/minimal.csv",
                "destinations": []
                }

    assert isinstance(CsvReader.necessary_config(node_config), set)
    assert len(CsvReader.necessary_config(node_config)) > 0


def test_kwargs():
    config = {
        "implementation_config":{
            "reader_config": {
            "csv_reader": {
                "class": "CsvReader",
                "filename": "test/minimal.csv",
                "kwargs": {
                    "header": None,
                    "sep": ":"
                },
                "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = CsvReader(configuration, "csv_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    df = data_object.get('csv_reader', rtype=DataObjectResponseType.VALUE.value)
    assert df is not None
    assert df.shape == (3, 1)



"""Transform data by filtering in data using filtering operations 

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import operator
import logging
import numpy as np
import pandas as pd

from primrose.base.transformer import AbstractTransformer

class FilterByPandasExpression(AbstractTransformer):
    """Applies filters to data as defined in feature_filters"""

    def __init__(self, feature_filters):
        """initialize filter with a list of feature_filters"""
        self.feature_filters = feature_filters

    def fit(self, data):
        """fit data, here just passing

        Args:
            data (object): some data

        """
        pass

    def transform(self, data):
        """Applies filters to data as defined in feature_filters. This is neccessary so we can filter rows in one reader
        based on information from another, and therefore has to be applied after the combiner step.

        The filters can operate on a single column with a fixed set of operations and a static value:

        fixed operations: `==`, `!=`, `<>`, `<`, `<=`, `>`, `>=`

        The feature_filters object should be structured as a list of lists:

        feature_filters: `[[column, operation, static value], [column, operation, static value]]`

        example: `[["number_of_members", "<", 1000]]` for filtering all rows with number_of_members less than 1000

        Args:
            data (dict): dictionary with dataframes from all readers
            data_key (str): key to pull the dataframe from within the data object
            feature_filters (list): list of lists with columns and operators to filter on
            instance_name (str): name of this pipeline instance

        Returns: 
            dataframe with filtered data

        Raises:
            Exception if not a pandas dataframe, operation not supported, or column name not recognized

        """
        if not isinstance(data, pd.DataFrame):
            raise Exception("Data is not a pandas DataFrame")

        ops = {
            "==": operator.eq,
            "!=": operator.ne,
            "<>": operator.ne,
            "<": operator.lt,
            "<=": operator.le,
            ">": operator.gt,
            ">=": operator.ge
        }

        if len(self.feature_filters) == 0:
            logging.info("no filters found; returning combined_data as filtered_data")
            return data

        else:

            filters = []

            for col, op, val in self.feature_filters:

                if not col in data.columns:
                    raise Exception("Unrecognized filter column '" + str(col) + "'")

                if str(op) not in ops:
                    raise Exception("Unsupported filter operation '" + str(op) + "'")

                filters.append(ops[op](data[col], val))

            filtered_data = data[np.all(filters, axis=0)]
            filtered_data = filtered_data.reset_index(drop=True)

            logging.info('Filtered out %d rows' % (len(data) - len(filtered_data)))

        return filtered_data




import pytest
import os
import dill
from primrose.configuration.configuration import Configuration
from primrose.writers.dill_writer import DillWriter
from primrose.data_object import DataObject
from primrose.readers.csv_reader import CsvReader

@pytest.fixture()
def config():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_file_writer']
                }
            },
            "writer_config": {
                "recipe_file_writer": {
                    "class": "FileWriter",
                    "key": "test_data",
                    "dir": "cache",
                    "filename": "unittest_file_writer.dill"
                }
            }
        }
    }
    return config


def test_necessary_config(config):
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    writer = DillWriter(configuration, 'recipe_file_writer')
    node_config = {
                    "class": "FileWriter",
                    "key": "test_data",
                    "dir": "cache",
                    "filename": "unittest_file_writer.dill"
                }
    assert isinstance(writer.necessary_config(node_config), set)
    assert len(writer.necessary_config(node_config)) > 0


def test_init_ok(config):

    test_data_string = "some test data"

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    requestor = CsvReader(configuration, 'csv_reader')

    data_object.add(requestor, test_data_string, 'test_data')

    writer = DillWriter(configuration, 'recipe_file_writer')

    c = configuration.config_for_instance('recipe_file_writer')
    filename = c['dir'] + os.path.sep + c['filename']

    #clean out test file location
    if os.path.exists(filename):
        os.remove(filename)

    data_object, terminate = writer.run(data_object)

    assert not terminate

    assert os.path.exists(filename)

    read_data = dill.load(open(filename, 'rb'))

    assert test_data_string == read_data


"""Abstract module to traverse a DAG

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from abc import ABC, abstractmethod

class DagTraverser(ABC):
    """Abstract module to traverse a DAG"""

    def __init__(self, configuration):
        """instantiate the traverser

        Args:
            configuration (Configuration): Configuration instance

        """
        self.configuration = configuration

    @abstractmethod
    def run_section_by_section(self):
        """Do we want go through section by section?

        Note:
            suppose we had 5 sections and we asked to run just 2: section1 and section2.
            Is it important that we run section1 and then section 2 (if so, return True)
            or just that we run all the nodes from section1 and section2 together in some
            desired order (return False)

        Returns:
            result (Boolean)

        """

    @abstractmethod
    def traversal_list(self):
        """get list of nodes to traverse

        Returns:
            sequence (list): list of node name in the order that they should be run

        """
        pass # pragma: no cover



import pytest
import pandas as pd
from sklearn import preprocessing

from primrose.transformers.categoricals import ImplicitCategoricalTransform


@pytest.fixture()
def data():
    return pd.DataFrame({'one': ['a', 'b'], 'two': ['c', 'c']})


def test_implicit_cat_fit(data):

    ict = ImplicitCategoricalTransform('two')

    out = ict.fit(data)

    assert isinstance(ict.target_encoder, preprocessing.LabelEncoder)
    assert isinstance(ict._encoder['one'], preprocessing.LabelEncoder)
    assert isinstance(ict._encoder['two'], preprocessing.LabelEncoder)


def test_implicit_cat_transform(data):

    ict = ImplicitCategoricalTransform('two')

    out = ict.fit(data)

    out = ict.transform(out)

    assert set(out['one'].unique()) == set([0, 1])



import pytest
from primrose.models.sklearn_model import SklearnModel
from sklearn.linear_model import LinearRegression

def test__instantiate_model():
    model = SklearnModel._instantiate_model("linear_model.LinearRegression", args=None)
    assert isinstance(model, LinearRegression)

    with pytest.raises(Exception) as e:
        SklearnModel._instantiate_model("linear_model.junk", args=None)
    assert "Sklearn model junk not found in linear_model module" in str(e)






import pytest
from primrose.node_factory import NodeFactory
from testfixtures import LogCapture

from primrose.readers.csv_reader import CsvReader
from primrose.writers.dill_writer import DillWriter

def test_register_module_classes():

    with LogCapture() as l:
        NodeFactory().register_module_classes(__name__)
    l.check(
        ('root','INFO','Discovered class CsvReader (<class '
            "'primrose.readers.csv_reader.CsvReader'>)"),
        ('root',
            'DEBUG',
            "Registered CsvReader : <class 'primrose.readers.csv_reader.CsvReader'>"),
        ('root',
            'INFO',
            'Discovered class DillWriter (<class '
            "'primrose.writers.dill_writer.DillWriter'>)"),
        ('root',
            'DEBUG',
            "Registered DillWriter : <class 'primrose.writers.dill_writer.DillWriter'>")
    )



import pytest
import pandas as pd
from primrose.transformers.combine import left_merge_dataframe_on_validated_join_keys
from testfixtures import LogCapture
from primrose.transformers.combine import LeftJoinDataCombiner

def test_left_merge_dataframe_on_validated_join_keys():
    corpus = pd.read_csv("test/minimal.csv")
    data_out = left_merge_dataframe_on_validated_join_keys(corpus, right_df=None, join_keys=None)
    assert data_out is corpus

def test_left_merge_dataframe_on_validated_join_keys2():
    corpus = pd.read_csv("test/minimal.csv")

    with pytest.raises(Exception) as e:
        left_merge_dataframe_on_validated_join_keys(corpus, right_df=corpus, join_keys=['JUNK'])
    assert "Join key JUNK not in left Index(['first', 'last'], dtype='object'). Aborting merge." in str(e)

def test_left_merge_dataframe_on_validated_join_keys3():
    corpus = pd.read_csv("test/minimal.csv")
    right_df = pd.read_csv("test/merge_right.csv")

    with pytest.raises(Exception) as e:
        left_merge_dataframe_on_validated_join_keys(corpus, right_df=right_df, join_keys=['first'])
    assert "Join key first not in right Index(['col1', 'col2'], dtype='object'). Aborting merge." in str(e)

def test_left_merge_dataframe_on_validated_join_keys4():
    right_df = pd.read_csv("test/minimal.csv")
    left_df = pd.read_csv("test/merge_right2.csv")

    with pytest.raises(Exception) as e:
        left_merge_dataframe_on_validated_join_keys(left_df, right_df=right_df, join_keys=['first'])
    assert "Cannot cast join key first as int64" in str(e)

def test_left_merge_dataframe_on_validated_join_keys5():
    left_df = pd.read_csv("test/minimal.csv")
    right_df = pd.read_csv("test/merge_right3.csv")

    out = left_merge_dataframe_on_validated_join_keys(left_df, right_df=right_df, join_keys=['first'])
    assert out.shape[0] == 2

    assert list(out.T.to_dict().values())[0] == {'first': 'joe', 'last': 'doe', 'age': 47}
    assert list(out.T.to_dict().values())[1] == {'first': 'mary', 'last': 'poppins', 'age': 42}

def test_left_merge_dataframe_on_validated_join_keys_fanout():
    left_df = pd.read_csv("test/minimal.csv")
    right_df = pd.read_csv("test/merge_right4.csv")


    with LogCapture() as l:
        out = left_merge_dataframe_on_validated_join_keys(left_df, right_df=right_df, join_keys=['first'])
    l.check(
        ('root', 'WARNING', 'Merge increased data size by 1 rows.')
    )

    assert out.shape[0] == 3

    assert list(out.T.to_dict().values())[0] == {'first': 'joe', 'last': 'doe', 'age': 47}
    assert list(out.T.to_dict().values())[1] == {'first': 'joe', 'last': 'doe', 'age': 48}
    assert list(out.T.to_dict().values())[2] == {'first': 'mary', 'last': 'poppins', 'age': 42}

def test_LeftJoinDataCombiner():
    combiner = LeftJoinDataCombiner(["first"])
    # does nothing
    combiner.fit(None)

    corpus = pd.read_csv("test/minimal.csv")

    with LogCapture() as l:
        out = combiner.transform([corpus])
    l.check(
        ('root', 'WARNING', 'Combiner needs at least two reader inputs, passing unchanged data.')
    )
    assert out == [corpus]

def test_LeftJoinDataCombiner2():
    with pytest.raises(Exception) as e:
         LeftJoinDataCombiner(["first"]).transform("JUNK")
    assert "In this transformer, data needs to be a list of dataframes" in str(e)

def test_LeftJoinDataCombiner3():
    with pytest.raises(Exception) as e:
         LeftJoinDataCombiner(["first"]).transform(["JUNK", "JUNK2"])
    assert "LeftJoinDataCombiner must operate on an iterable of pandas.DataFrame objects." in str(e)

def test_LeftJoinDataCombiner4():
    left_df = pd.read_csv("test/minimal.csv")
    right_df = pd.read_csv("test/merge_right3.csv")
    data_list = [left_df, right_df]
    out = LeftJoinDataCombiner(["first"]).transform(data_list)
    assert out.shape[0] == 2

    assert list(out.T.to_dict().values())[0] == {'first': 'joe', 'last': 'doe', 'age': 47}
    assert list(out.T.to_dict().values())[1] == {'first': 'mary', 'last': 'poppins', 'age': 42}





import pytest
import sys
import os
import jstyleson
from primrose.configuration.configuration_dag import ConfigurationDag
from primrose.configuration.configuration import Configuration
from primrose.configuration.configuration import OperationType
from primrose.node_factory import NodeFactory
from primrose.base.node import AbstractNode
from primrose.dag.config_layer_traverser import ConfigLayerTraverser

def test_check_node_exists():
    with pytest.raises(Exception) as e:
        ConfigurationDag.check_node_exists(["a","b","c"],"d")
    assert 'Destination d does not exist' in str(e)

def test_check_cycles_OK():
    config = {
          "implementation_config":{
            "reader_config":{
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }

    dag = ConfigurationDag(config['implementation_config'])
    dag.create_dag()
    assert dag.check_for_cycles() is None

def test_check_cycles_bad():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": ["csv_reader"]
                }
            }
        }
    }

    dag = ConfigurationDag(config['implementation_config'])
    dag.create_dag()
    with pytest.raises(Exception) as e:
        dag.check_for_cycles()
    assert "Cycle(s) found:" in str(e)
    # it turns out cycle finding code is stochastic so the order of the nodes in the found cycle is not stable. Thus, we look for initial phrase only
    # assert "Cycle(s) found: [('corpus_pipeline', 'recipe_name_model'), ('recipe_name_model', 'corpus_pipeline')]" in str(e)

def test_check_cycles_bad2():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": []
                }
            }
        }
    }

    dag = ConfigurationDag(config['implementation_config'])
    dag.create_dag()
    with pytest.raises(Exception) as e:
        dag.check_connected_components()
    assert "Found multiple connected components:" in str(e)

def test_type_bad():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": [1234]
                }
            },
            "writer_config": {
                1234: {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": []
                }
            }
        }
    }

    dag = ConfigurationDag(config['implementation_config'])
    with pytest.raises(Exception) as e:
        dag.create_dag()
    assert "Unrecognized destination type: 1234" in str(e)

def test_type_bad2():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['doesnotexist']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": []
                }
            }
        }
    }

    dag = ConfigurationDag(config['implementation_config'])
    with pytest.raises(Exception) as e:
        dag.create_dag()
    assert "Did not find doesnotexist destination in" in str(e)

@pytest.mark.optional
@pytest.mark.plotting
def test_plot_dag():

    class TestPostprocess(AbstractNode):
        @staticmethod
        def necessary_config(node_config):
            return set(['key1', 'key2'])
        def run(self, data_object):
            return data_object
    NodeFactory().register("TestPostprocess", TestPostprocess)

    class Testpipeline(AbstractNode):
        #def __init__(self, configuration, instance_name):
        #    pass
        @staticmethod
        def necessary_config(node_config):
            return set([])
        def run(self, data_object):
            return data_object
    NodeFactory().register("Testpipeline", Testpipeline)
    class TestCleanup(AbstractNode):
        @staticmethod
        def necessary_config(node_config):
            return set([])
        def run(self, data_object):
            return data_object
    NodeFactory().register("TestCleanup", TestCleanup)

    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "some/path/to/file",
                    "destinations": ["pipeline1"]
                }
            },
            "pipeline_config": {
                "pipeline1": {
                    "class": "Testpipeline",
                    "destinations": ["decision_tree_model"]
                }
            },
            "model_config": {
                "decision_tree_model": {
                    "class": "SklearnClassifierModel",
                    "model_parameters": {},
                    "sklearn_classifier_name": "tree.DecisionTreeClassifier",
                    "grid_search_scoring": "roc_auc",
                    "cv_folds": 3,
                    "mode": "predict",
                    "destinations": ["nodename"]
                }
            },
            "postprocess_config": {
                "nodename": {
                    "class": "TestPostprocess",
                    "key1":"val1",
                    "key2":"val2",
                    "destinations": ["write_output"]
                }
            },
            "writer_config": {
                "write_output": {
                    "class": "CsvWriter",
                    "key": "read_data",
                    "dir": "cache",
                    "filename": "some/path/to/file.csv",
                    "destinations": ["donothingsuccess"]
                }
            },
            "cleanup_config": {
                "donothingsuccess": {
                    "class": "TestCleanup",
                }
            }
        }
    }

    cfilename = "test/test_dag_plotting.json"
    with open(cfilename, 'w') as f:
        jstyleson.dump(config, f)
    config = Configuration(config_location=cfilename)
    dag = config.dag
    dag.create_dag()

    filename = "test/test_dag_plotting.png"
    if os.path.exists(filename):
        os.remove(filename)

    dag.plot_dag(filename, traverser=ConfigLayerTraverser(config))

    assert os.path.exists(filename)

    if os.path.exists(cfilename):
        os.remove(cfilename)

    if os.path.exists(filename):
        os.remove(filename)

def test_upstream_keys1():
    config = {
        "implementation_config":{
            "reader_config":{
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    dag = ConfigurationDag(config['implementation_config'])
    dag.create_dag()
    ukeys = dag.upstream_keys('csv_writer')
    assert list(ukeys) == ['csv_reader']

    with pytest.raises(Exception) as e:
        dag.upstream_keys('doesnotexist')
    assert "Node not found in the DAG:" in str(e)

def test_upstream_typed_keys():
    config = {
        "implementation_config":{
            "reader_config":{
                "csv_reader1": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                },
                "csv_reader2": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    dag = ConfigurationDag(config['implementation_config'])
    dag.create_dag()
    ukeys = dag.upstream_typed_keys('csv_writer')
    assert ukeys == {'csv_reader1': 'reader_config', 'csv_reader2': 'reader_config'}

def test_starting_nodes():
    #here, we also test descendents()

    config = {
        "metadata": {},

        "implementation_config":{
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["decision_tree_model"]
                },
                "read_data2": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["decision_tree_model"]
                }
            },
            "model_config": {
                "decision_tree_model": {
                    "class": "SklearnClassifierModel",
                    "model_parameters": {},
                    "sklearn_classifier_name": "tree.DecisionTreeClassifier",
                    "grid_search_scoring": "roc_auc",
                    "cv_folds": 3,
                    "mode": "predict",
                    "destinations": ["write_output"]
                }
            },
            "writer_config": {
                "write_output": {
                    "class": "CsvWriter",
                    "key": "predictions",
                    "dir": "cache",
                    "filename": "hello_world_predictions.csv"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    nodes = configuration.dag.starting_nodes()
    assert nodes == set(["read_data", "read_data2"])

    downstream_nodes = configuration.dag.descendents("read_data2")
    assert downstream_nodes == set(["decision_tree_model", "write_output"])

def test_nodes_of_type():
    config = {
        "metadata": {},

        "implementation_config":{
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["decision_tree_model"]
                },
                "read_data2": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["decision_tree_model"]
                }
            },
            "model_config": {
                "decision_tree_model": {
                    "class": "SklearnClassifierModel",
                    "mode": "predict",
                    "sklearn_classifier_name": "tree.DecisionTreeClassifier",
                    "grid_search_scoring": "roc_auc",
                    "cv_folds": 3,
                    "model_parameters": {},
                    "destinations": ["write_output"]
                }
            },
            "writer_config": {
                "write_output": {
                    "class": "CsvWriter",
                    "key": "predictions",
                    "dir": "cache",
                    "filename": "hello_world_predictions.csv"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    nodes = configuration.dag.nodes_of_type(OperationType.reader)
    assert nodes == set(["read_data", "read_data2"])

    nodes = configuration.dag.nodes_of_type(OperationType.pipeline)
    assert nodes == set([])

    nodes = configuration.dag.upstream_nodes_of_type("write_output", OperationType.reader)
    assert nodes == set(["read_data", "read_data2"])

    nodes = configuration.dag.upstream_nodes_of_type("write_output", OperationType.cleanup)
    assert nodes == set([])


def test_paths():
    config = {
        "metadata": {},

        "implementation_config":{
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["read_data2", "decision_tree_model"]
                },
                "read_data2": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["decision_tree_model"]
                }
            },
            "model_config": {
                "decision_tree_model": {
                    "class": "SklearnClassifierModel",
                    "mode": "train",
                    "sklearn_classifier_name": "tree.DecisionTreeClassifier",
                    "grid_search_scoring": "roc_auc",
                    "cv_folds": 3,
                    "model_parameters": {},
                    "destinations": ["write_output"]
                }
            },
            "writer_config": {
                "write_output": {
                    "class": "CsvWriter",
                    "key": "predictions",
                    "dir": "cache",
                    "filename": "hello_world_predictions.csv"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    paths = list(configuration.dag.paths("read_data", "write_output"))
    assert len(paths) == 2
    assert paths[0] == ['read_data', 'read_data2', 'decision_tree_model', 'write_output']
    assert paths[1] == ['read_data', 'decision_tree_model', 'write_output']

    paths = configuration.dag.paths("write_output", "decision_tree_model")
    assert not paths




"""e
    Module with all tests for the slack integration implementations
    author(s): Parul Laul (parul.laul@ww.com)
"""
import unittest.mock as mock
import pytest
import logging

from primrose.notification_utils import SlackClient, get_notification_client


def test_post():
    client_mock = mock.Mock(return_value=mock.Mock())

    # instantiate instance
    slack_instance = SlackClient(channel='some_channel', member_id='USomeUserID', token='some_token')
    slack_instance.client = client_mock.return_value

    slack_instance.post_message(message='test message')

    # check method is called
    slack_instance.client.chat_postMessage.assert_called_once_with(
        channel='some_channel',
        text='test message\n <@USomeUserID>'
    )

def test_get_notification_client():
    importlib_mock = mock.Mock()
    getattr_mock = mock.Mock()

    patches = {
        'importlib': importlib_mock,
        'getattr': getattr_mock
    }
    path = 'primrose.notification_utils'
    with mock.patch.multiple(path, **patches):
        client_params = {'some': 'params', 'client': 'needed', 'token': 'needed'}

        _ = get_notification_client(client_params)

    assert(importlib_mock.import_module.call_count == 1)
    assert(getattr_mock.call_count == 1)

def test_get_notification_client_exc(caplog):
    client_params = {'client': 'DoesNotExist'}
    with caplog.at_level(logging.ERROR):
        get_notification_client(client_params)
    assert("Are you sure DoesNotExist is in <module 'primrose.notification_utils'" in caplog.text)



"""Postgres helper

Author(s):
    Wassym Kalouache (wassym.kalouache@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
from primrose.readers.database_helper import get_env_val

try:
    import psycopg2
    HAS_PSYCOPG2 = True

except ImportError:
    HAS_PSYCOPG2 = False


class PostgresHelper():
    '''
        some utility methods for connecting to postgres
    '''

    @staticmethod
    def extract_postgres_credentials():
        """extract PostgreSQL credentials from config

        Returns:
            (tuple): tuple containing: \
                host (str): host \
                port (int): port \
                database (str): database name \
                username (str): username \
                password (str): password \

        """
        host = get_env_val("POSTGRES_HOST")
        port = int(get_env_val("POSTGRES_PORT"))
        database = get_env_val("POSTGRES_DB")
        username = get_env_val("POSTGRES_USER")
        password = get_env_val("POSTGRES_PASS")
        return (host, port, username, password, database)

    @staticmethod
    def create_db_connection():
        """authenticate with postgres database

            Returns:
                db (connection): postgres db object

        """
        if not HAS_PSYCOPG2:
            raise ImportError("psycopg2 is necessary to establish connection")

        host, port, username, password, database = PostgresHelper.extract_postgres_credentials()
        conn = psycopg2.connect(dbname=database, user=username, password=password, host=host, port=port, sslmode='require')
        return conn




"""write some data to a local file.

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
import logging
from primrose.writers.abstract_file_writer import AbstractFileWriter
from primrose.data_object import DataObjectResponseType

class FileWriter(AbstractFileWriter):
    ''' write some data object to a local file '''

    def run(self, data_object):
        """write some data to a local file

        Note:
            this will write object under config['key'] in data_object

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        filename = self.node_config['dir'] + os.path.sep + self.node_config['filename']
        data_key = self.node_config['key']
        logging.info("Saving %s data to %s", data_key, filename)
        data_to_write = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.VALUE.value)

        with open(filename, 'w') as f:
            f.write(data_to_write)
        f.close()

        terminate = False
        return data_object, terminate




import pytest
from primrose.configuration.util import *

def test_ConfigurationSectionType():
    assert ConfigurationSectionType.values() == ['metadata', 'implementation_config']

def test_OperationType_names():
    assert OperationType.names() == ['reader', 'pipeline', 'model', 'postprocess', 'writer', 'dataviz', 'cleanup']

def test_OperationType_values():
    assert OperationType.values() == ['reader_config', 'pipeline_config', 'model_config', 'postprocess_config', 'writer_config', 'dataviz_config', 'cleanup_config']

def test_values_to_names():
    assert OperationType.values_to_names() == {'reader_config': 'reader', 'pipeline_config': 'pipeline', 'model_config':'model', 'postprocess_config':'postprocess', 'writer_config':'writer', 'dataviz_config':'dataviz', 'cleanup_config':'cleanup'}



'''Top level notion of a node in the graph

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

'''
from abc import ABC, abstractmethod

class AbstractNode(ABC):

    def __init__(self, configuration, instance_name):
        """

        Args:
            configuration (Configuration): configuration object defined in primrose/Configuration with validated inputs
                from the result of necessary_config, all inputs are described in that method

            instance_name (str): how the code knows where it is from

        """
        self.configuration = configuration
        self.instance_name = instance_name
        self.node_config = configuration.config_for_instance(instance_name)

    @staticmethod
    @abstractmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        After adding this list, validation automatically occurs before instantiation in the
        configuration

        Returns:
            set of keys necessary to run implementation

        """
        return set() # pragma: no cover

    @abstractmethod
    def run(self, data_object):
        """
            run the node. For a reader, that means read, for a writer that means write etc.

        Args:
            data_object (DataObject): DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        pass # pragma: no cover



"""Module to run preprocessing using SKlearn preprocessors

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import importlib
from primrose.base.transformer_sequence import TransformerSequence
from primrose.transformers.sklearn_preprocessing_transformer import SklearnPreprocessingTransformer
from primrose.data_object import DataObjectResponseType
from primrose.pipelines.train_test_split import TrainTestSplit

class SklearnPreprocessingPipeline(TrainTestSplit):

    @staticmethod
    def necessary_config(node_config):
        """Return the necessary configuration keys for the SklearnPreprocessingPipeline object

        Returns:
            set of keys

        """
        return set(['operations']).union(TrainTestSplit.necessary_config(node_config))

    def init_pipeline(self):
        """create the pipeline's TransformerSequence

        Returns:
            a TransformerSequence

        """
        ts = TransformerSequence()

        for operation in self.node_config["operations"]:

            args = operation.get('args', None)
            columns = operation.get('columns', None)
            
            p = self._instantiate_preprocessor(operation['class'], args, columns)
            ts.add(p)

        return ts

    @staticmethod
    def _instantiate_preprocessor(classname, args=None, columns=None):
        """Import and validate user-defined sklearn preprocessor
        
        Returns:
            SklearnPreprocessingTransformer
        
        """

        sk_module_name, sk_transformer_name = classname.split('.')
        sk_module = importlib.import_module('sklearn.{}'.format(sk_module_name))

        try:
            t = getattr(sk_module, sk_transformer_name)
        except AttributeError:
            raise Exception('Preprocessor {} not found in {} module'.format(sk_transformer_name, sk_module_name))

        if args:
            return SklearnPreprocessingTransformer(t(**args), columns)

        return SklearnPreprocessingTransformer(t(), columns)





"""
    Module with AbstractSuccess implementation.
    Provides slack notification after job completion

    author(s): Parul Laul (parul.laul@ww.com)

"""

import importlib
import os
import logging

from primrose.base.success import AbstractSuccess
from primrose.data_object import DataObjectResponseType

from primrose.notification_utils import get_notification_client



def get_client_params(params: dict):
    """
        Creates the parameter dictionary to be read by the
        `get_notification_client` method.

        Args:
            params (`dict`): The parameter dictionary where keys are
                - the primrose `client` (required key),
                - neccessary arguments to instantiate the client
                - the `message` if used in the `implementation` section of the DAG.

                Values can be explicitly or stored as environment variables.
                Environment variables are stored as {CLIENT_NAME}_{CLIENT_KEY}.

                For example, using the built-in SlackClient, environment variables
                will be stored as:
                    SLACKCLIENT_TOKEN="some-token"
                    SLACKCLIENT_CHANNEL="some-channel"
                    SLACKCLIENT_MEMBER_ID="USomeUserID"

        Returns:
            A `dict` with the key-value pairs necessary to be read by the
            `get_notification_client` method.

        Example:
            >>> node_config = {
                    "client": "SlackClient",
                    "message": "starting job...",
                }
            >>> get_client_params(node_config)
            {'client': 'SlackClient',
            'channel': 'some-channel',
            'message': 'starting job...',
            'token': 'some-token',
            'member_id': 'USomeUserID'
            }

    """
    client = params['client'].upper()
    env_var = {
        k.split(f'{client}_')[-1].lower(): v for k, v in os.environ.items()
        if client.upper() in k
        and k.split(f'{client}_')[-1].lower() not in params.keys() # config params take precedence
    }

    # combine environment variables and params
    return {**params, **env_var}


class ClientNotification(AbstractSuccess):
    """Outputs success notification using specified client after job completion."""

    def __init__(self, configuration, instance_name):
        """Outputs success message to notify user of DAG completion.

        Args:
            configuration: configuration object from src/configuration
            instance_name: name key to use for this configuration

        Returns:
            None

        """
        super().__init__(configuration, instance_name)
        self.message = self.node_config.get("message", "SUCCESS! DAG Completed")
        self.use_configuration_file_message = self.node_config.get("use_configuration_file_message", True)
        self.node_name = self.node_config.get("node_name", "")
        self.message_key = self.node_config.get("message_key", "")

        # read from config dict or environment variables
        self.client = get_notification_client(params=get_client_params(self.node_config))

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the ClientNotification object

            Notes:
                client (`str`): The client object to be instantiated (ex. 'SlackClient')
                token (`str`): The token used to initialize the client

        """
        config_params = set(["client", "token"])
        return config_params.union(AbstractSuccess.necessary_config(node_config))

    @staticmethod
    def optional_config(node_config):
        """Returns the optional configuration keys

        Args:
            node_config (dict): set of parameters for the node

        optional keys:
            use_configuration_file_message(bool): True if you want the slack client to send message from
            configuration file. This is the default behavior. If you set it to false
            pass the next two parameters to send alternate message passed from another node.

            node_name(str): the name of the node where this module will find the message to
            post

            message_key(str): the key in the node that should be used for getting the message
        Returns:
            Set of optional keys for the pipeline object
        """
        return set(["use_config_message", "node_name", "message_key",])

    def run(self, data_object):
        """
            Run job to post slack message

            Args:
                data_object (`DataObject`): primrose data object instance
                terminate (`bool`). Should we terminate the DAG? True or False

        """

        if not self.use_configuration_file_message:
            logging.info("Posting message passed by a node")

            if self.node_name and self.message_key:
                upstream_data = data_object.get_upstream_data(
                    self.instance_name, pop_data=False, rtype=DataObjectResponseType.INSTANCE_KEY_VALUE.value,
                )

                self.message = upstream_data[self.node_name][self.message_key]
            else:
                logging.info(
                    """Do not have adequate information to retrieve message from
                                an upstream node; reverting to default behavior.
                                If you want to post message from an upstream node, provide
                                values for node_name and message_key """
                )
    
        # execute response
        _ = self.client.post_message(message=self.message)

        terminate = False
        return data_object, terminate




import pytest
import pandas as pd
from primrose.transformers.categoricals import ExplicitCategoricalTransform

def test__process_transformations():
    categoricals = {
        "currency": {
            "transformations": [
                "{x}[{x}=='USD'] = 1",
                "{x}[{x}=='EUR'] = 2"
            ],
            "rename": "money",
            "to_numeric": True
        }
    }
    x = "data['{}']".format("currency")
    input_data = categoricals["currency"]

    data = pd.DataFrame(data={"currency": ['EUR', 'USD', 'USD']})

    ExplicitCategoricalTransform._process_transformations(data, input_data, "currency", x)

    assert list(data.currency) == [2, 1, 1]
    assert "money" not in list(data.columns)

    data, new_name = ExplicitCategoricalTransform._process_rename(data, input_data, "currency")

    assert "money" in list(data.columns)
    assert list(data.money) == [2, 1, 1]

    assert str(data.dtypes['money']) == "object"

    data = ExplicitCategoricalTransform._process_numeric(data, input_data, new_name)

    assert str(data.dtypes['money']) == "int64"

def test__process_numeric():
    input_data = {
        "to_numeric": True
    }
    data = pd.DataFrame(data={"currency": ['EUR', 'USD', 'USD']})
    data = ExplicitCategoricalTransform._process_numeric(data, input_data, "currency")

    assert list(data.currency) == 3*[ExplicitCategoricalTransform.DEFAULT_NUMERIC]

def test__process_numeric_no_config_key():
    input_data = {}
    data = pd.DataFrame(data={"currency": ['EUR', 'USD', 'USD']})
    data = ExplicitCategoricalTransform._process_numeric(data, input_data, "currency")

    assert data is not None

def test__process_rename():
    input_data = {
        "transformations": [
            "{x}[{x}=='USD'] = 1",
            "{x}[{x}=='EUR'] = 2"
        ]
    }
    data = pd.DataFrame(data={"currency": ['EUR', 'USD', 'USD']})
    data, new_name = ExplicitCategoricalTransform._process_rename(data, input_data, "currency")
    assert new_name == "currency"

def test_transform():
    categoricals = {
        "currency": {
            "transformations": [
                "{x}[{x}=='USD'] = 1",
                "{x}[{x}=='EUR'] = 2"
            ],
            "rename": "money",
            "to_numeric": True
        }
    }
    data = pd.DataFrame(data={"currency": ['EUR', 'USD', 'USD']})
    t = ExplicitCategoricalTransform(categoricals)

    #does nothing
    t.fit(data)

    data_out = t.transform(data)
    assert list(data_out.money) == [2, 1, 1]
    assert str(data_out.dtypes['money']) == "int64"



"""Custom imputer to replace values within pandas dataframe columns with config-specified imputation schemes

Author(s):
    Reka Daniel-Weiner (reka.danielweiner@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import pandas as pd
import logging
from primrose.base.transformer import AbstractTransformer

class ColumnSpecificImpute(AbstractTransformer):
    """Transform config specified columns NULL values into zero, mean, median, mode, inf or negative inf"""

    def __init__(self, columns_to_zero, columns_to_mean, columns_to_median, columns_to_mode,
                columns_to_infinity, columns_to_neg_infinity):
        """Transform config specified columns NULL values into zero, mean, median, mode, inf or negative inf

        Args:
            columns_to_zero (list): list of columns to impute zeros
            columns_to_mean (list): list of columns to impute means
            columns_to_median (list): list of columns to impute medians
            columns_to_mode (list): list of columns to impute modes
            columns_to_infinity (list): list of columns to impute to large value (999999999)
            columns_to_neg_infinity (list): list of columns to impute to large negative value (-999999999)

        Returns:
            nothing. Side effect to set list of columns to set to mean, 0, median etc

        """
        self.columns_to_zero = columns_to_zero
        self.columns_to_mean = columns_to_mean
        self.columns_to_median = columns_to_median
        self.columns_to_mode = columns_to_mode
        self.columns_to_infinity = columns_to_infinity
        self.columns_to_neg_infinity = columns_to_neg_infinity
        self.encoder = None

    def fit(self, data):
        """Fit encoder imputation values to dataframe metrics
        Create a dictionary of column name to imputed values. These values might be straight constants or might
        be a function of the complete column's value, such as mode or median

        Args:
            data (pandas data frame): a data frame

        Returns:
            Nothing. Updates internal dictionary of column name to imputed value

        Raises:
            Exception if a column appears in multiple lists, or if column not recognized

        """

        self.encoder = {}
        columns_so_far = set()

        for cols in [self.columns_to_zero,
                    self.columns_to_mean,
                    self.columns_to_median,
                    self.columns_to_mode,
                    self.columns_to_infinity,
                    self.columns_to_neg_infinity]:

            # does column exist?
            for col in cols:
                if not col in data.columns:
                    raise Exception("Unrecognized impute column '" + str(col) + "'")

            # is it in some previous list?
            in_common = columns_so_far.intersection(set(cols))
            if in_common:
                raise Exception("There are columns in multiple lists "+ str(in_common))
            columns_so_far = columns_so_far.union(set(cols))

        logging.info('Specifying columns to impute 0')
        [self.encoder.setdefault(col, 0) for col in self.columns_to_zero]

        logging.info('Specifying columns to impute median')
        [self.encoder.setdefault(col, data[col].median()) for col in self.columns_to_median]

        logging.info('Specifying columns to impute mean')
        [self.encoder.setdefault(col, data[col].mean()) for col in self.columns_to_mean]

        logging.info('Specifying columns to impute large values')
        [self.encoder.setdefault(col, 999999999.) for col in self.columns_to_infinity]

        logging.info('Specifying columns to impute large negative values')
        [self.encoder.setdefault(col, -999999999.) for col in self.columns_to_neg_infinity]

        logging.info('Specifying columns to impute mode')
        for col in self.columns_to_mode:
            logging.info('imputing {}'.format(col))
            try:
                if pd.notnull(data[col].mode().values[0]):
                    col_mode = data[col].mode().values[0]
                else:
                    col_mode = 0
            except Exception:
                col_mode = 0

            self.encoder[col] = col_mode

    def transform(self, data):
        """Impute columns in data according to the imputations fit by self.fit

        Args:
            data (dataframe)

        Returns:
            data (dataframe)

        Raises:
            Exception if train is not called before transfoorm

        """

        if self.encoder is None:
            raise Exception('ColumnSpecificImpute must train imputations with fit before calling transform.')

        for col, val in self.encoder.items():
            data[col] = data[col].fillna(val)

        return data
